from dagster.core.storage.object_store import DEFAULT_SERIALIZATION_STRATEGY
from dagster_aws.s3 import S3ObjectStore


def test_s3_object_store(bucket, caplog):
    key = "foo"

    s3_obj_store = S3ObjectStore(bucket.name)
    res_key = s3_obj_store.set_object(key, True, DEFAULT_SERIALIZATION_STRATEGY)
    assert res_key == "s3://{s3_bucket}/{key}".format(s3_bucket=bucket.name, key=key)

    s3_obj_store.set_object(key, True, DEFAULT_SERIALIZATION_STRATEGY)
    assert "Removing existing S3 key" in caplog.text

    assert s3_obj_store.has_object(key)
    assert s3_obj_store.get_object(key, DEFAULT_SERIALIZATION_STRATEGY)[0] == True

    s3_obj_store.cp_object(key, "bar")
    assert s3_obj_store.has_object("bar")

    s3_obj_store.rm_object(key)
    assert not s3_obj_store.has_object(key)

    assert s3_obj_store.uri_for_key(key) == "s3://{s3_bucket}/{key}".format(
        s3_bucket=bucket.name, key=key
    )



import boto3
import pytest
from moto import mock_s3


@pytest.fixture
def s3():
    with mock_s3():
        yield boto3.resource("s3")


@pytest.fixture
def bucket(s3):  # pylint: disable=redefined-outer-name
    yield s3.create_bucket(Bucket="test-bucket")



import os
import uuid

import boto3
from dagster import (
    InputDefinition,
    Int,
    ModeDefinition,
    OutputDefinition,
    ResourceDefinition,
    configured,
    execute_pipeline,
    pipeline,
    solid,
)
from dagster.seven import mock
from dagster_aws.s3 import (
    S3FileHandle,
    S3FileManager,
    s3_file_manager,
    s3_plus_default_storage_defs,
)
from moto import mock_s3


def test_s3_file_manager_write():
    s3_mock = mock.MagicMock()
    file_manager = S3FileManager(s3_mock, "some-bucket", "some-key")

    foo_bytes = "foo".encode()

    file_handle = file_manager.write_data(foo_bytes)

    assert isinstance(file_handle, S3FileHandle)

    assert file_handle.s3_bucket == "some-bucket"
    assert file_handle.s3_key.startswith("some-key/")

    assert s3_mock.put_object.call_count == 1

    file_handle = file_manager.write_data(foo_bytes, ext="foo")

    assert isinstance(file_handle, S3FileHandle)

    assert file_handle.s3_bucket == "some-bucket"
    assert file_handle.s3_key.startswith("some-key/")
    assert file_handle.s3_key[-4:] == ".foo"

    assert s3_mock.put_object.call_count == 2


def test_s3_file_manager_read():
    state = {"called": 0}
    bar_bytes = "bar".encode()

    class S3Mock(mock.MagicMock):
        def download_file(self, *_args, **kwargs):
            state["called"] += 1
            assert state["called"] == 1
            state["bucket"] = kwargs.get("Bucket")
            state["key"] = kwargs.get("Key")
            file_name = kwargs.get("Filename")
            state["file_name"] = file_name
            with open(file_name, "wb") as ff:
                ff.write(bar_bytes)

    s3_mock = S3Mock()
    file_manager = S3FileManager(s3_mock, "some-bucket", "some-key")
    file_handle = S3FileHandle("some-bucket", "some-key/kdjfkjdkfjkd")
    with file_manager.read(file_handle) as file_obj:
        assert file_obj.read() == bar_bytes

    assert state["bucket"] == file_handle.s3_bucket
    assert state["key"] == file_handle.s3_key

    # read again. cached
    with file_manager.read(file_handle) as file_obj:
        assert file_obj.read() == bar_bytes

    assert os.path.exists(state["file_name"])

    file_manager.delete_local_temp()

    assert not os.path.exists(state["file_name"])


@mock_s3
def test_depends_on_s3_resource_intermediates():
    @solid(
        input_defs=[InputDefinition("num_one", Int), InputDefinition("num_two", Int)],
        output_defs=[OutputDefinition(Int)],
    )
    def add_numbers(_, num_one, num_two):
        return num_one + num_two

    # Uses mock S3
    s3 = boto3.client("s3")
    s3.create_bucket(Bucket="some-bucket")

    @pipeline(
        mode_defs=[
            ModeDefinition(
                system_storage_defs=s3_plus_default_storage_defs,
                resource_defs={"s3": ResourceDefinition.hardcoded_resource(s3)},
            )
        ]
    )
    def s3_internal_pipeline():
        return add_numbers()

    result = execute_pipeline(
        s3_internal_pipeline,
        run_config={
            "solids": {
                "add_numbers": {"inputs": {"num_one": {"value": 2}, "num_two": {"value": 4}}}
            },
            "storage": {"s3": {"config": {"s3_bucket": "some-bucket"}}},
        },
    )

    keys_in_bucket = [obj["Key"] for obj in s3.list_objects(Bucket="some-bucket")["Contents"]]
    assert result.success
    assert result.result_for_solid("add_numbers").output_value() == 6

    keys = set()
    for step_key, output_name in [("add_numbers.compute", "result")]:
        keys.add(create_s3_key(result.run_id, step_key, output_name))

    assert set(keys_in_bucket) == keys


def create_s3_key(run_id, step_key, output_name):
    return "dagster/storage/{run_id}/intermediates/{step_key}/{output_name}".format(
        run_id=run_id, step_key=step_key, output_name=output_name
    )


@mock_s3
def test_depends_on_s3_resource_file_manager():
    bar_bytes = "bar".encode()

    @solid(output_defs=[OutputDefinition(S3FileHandle)])
    def emit_file(context):
        return context.file_manager.write_data(bar_bytes)

    @solid(input_defs=[InputDefinition("file_handle", S3FileHandle)])
    def accept_file(context, file_handle):
        local_path = context.file_manager.copy_handle_to_local_temp(file_handle)
        assert isinstance(local_path, str)
        assert open(local_path, "rb").read() == bar_bytes

    # Uses mock S3
    s3 = boto3.client("s3")
    s3.create_bucket(Bucket="some-bucket")

    @pipeline(
        mode_defs=[
            ModeDefinition(
                system_storage_defs=s3_plus_default_storage_defs,
                resource_defs={"s3": ResourceDefinition.hardcoded_resource(s3)},
            )
        ]
    )
    def s3_file_manager_test():
        accept_file(emit_file())

    result = execute_pipeline(
        s3_file_manager_test,
        run_config={"storage": {"s3": {"config": {"s3_bucket": "some-bucket"}}}},
    )

    assert result.success

    keys_in_bucket = set([obj["Key"] for obj in s3.list_objects(Bucket="some-bucket")["Contents"]])

    for step_key, output_name in [
        ("emit_file.compute", "result"),
        ("accept_file.compute", "result"),
    ]:
        keys_in_bucket.remove(create_s3_key(result.run_id, step_key, output_name))

    assert len(keys_in_bucket) == 1

    file_key = list(keys_in_bucket)[0]
    comps = file_key.split("/")

    assert "/".join(comps[:-1]) == "dagster/storage/{run_id}/files".format(run_id=result.run_id)

    assert uuid.UUID(comps[-1])


@mock.patch("boto3.resource")
@mock.patch("dagster_aws.s3.resources.S3FileManager")
def test_s3_file_manger_resource(MockS3FileManager, mock_boto3_resource):
    did_it_run = dict(it_ran=False)

    resource_config = {
        "use_unsigned_session": True,
        "region_name": "us-west-1",
        "endpoint_url": "http://alternate-s3-host.io",
        "s3_bucket": "some-bucket",
        "s3_prefix": "some-prefix",
    }

    mock_s3_session = mock_boto3_resource.return_value.meta.client

    @solid(required_resource_keys={"file_manager"})
    def test_solid(context):
        # test that we got back a S3FileManager
        assert context.resources.file_manager == MockS3FileManager.return_value

        # make sure the file manager was initalized with the config we are supplying
        MockS3FileManager.assert_called_once_with(
            s3_session=mock_s3_session,
            s3_bucket=resource_config["s3_bucket"],
            s3_base_key=resource_config["s3_prefix"],
        )

        _, call_kwargs = mock_boto3_resource.call_args

        mock_boto3_resource.assert_called_once_with(
            "s3",
            region_name=resource_config["region_name"],
            endpoint_url=resource_config["endpoint_url"],
            use_ssl=True,
            config=call_kwargs["config"],
        )

        assert call_kwargs["config"].retries["max_attempts"] == 5

        did_it_run["it_ran"] = True

    @pipeline(
        mode_defs=[
            ModeDefinition(
                resource_defs={"file_manager": configured(s3_file_manager)(resource_config)},
            )
        ]
    )
    def test_pipeline():
        test_solid()

    execute_pipeline(test_pipeline)
    assert did_it_run["it_ran"]



import io
from collections import defaultdict

from botocore.exceptions import ClientError


def create_s3_fake_resource(buckets=None):
    """Create a mock S3 session for test."""
    return S3FakeSession(buckets=buckets)


class S3FakeSession(object):
    """Stateful mock of a boto3 s3 session for test.

    Wraps a ``mock.MagicMock``. Buckets are implemented using an in-memory dict.
    """

    def __init__(self, buckets=None):
        from dagster.seven import mock

        self.buckets = defaultdict(dict, buckets) if buckets else defaultdict(dict)
        self.mock_extras = mock.MagicMock()

    def head_bucket(self, Bucket, *args, **kwargs):  # pylint: disable=unused-argument
        self.mock_extras.head_bucket(*args, **kwargs)

    def head_object(self, Bucket, Key, *args, **kwargs):
        self.mock_extras.head_object(*args, **kwargs)
        return {"ContentLength": len(self.buckets.get(Bucket, {}).get(Key, b""))}

    def list_objects_v2(self, Bucket, Prefix, *args, **kwargs):
        self.mock_extras.list_objects_v2(*args, **kwargs)
        key = self.buckets.get(Bucket, {}).get(Prefix)
        if key:
            return {"KeyCount": 1, "Contents": [{"Key": key}], "IsTruncated": False}
        else:
            return {"KeyCount": 0, "Contents": [], "IsTruncated": False}

    def put_object(self, Bucket, Key, Body, *args, **kwargs):
        self.mock_extras.put_object(*args, **kwargs)
        self.buckets[Bucket][Key] = Body.read()

    def get_object(self, Bucket, Key, *args, **kwargs):
        if not self.has_object(Bucket, Key):
            raise ClientError({}, None)

        self.mock_extras.get_object(*args, **kwargs)
        return {"Body": self._get_byte_stream(Bucket, Key)}

    def upload_fileobj(self, fileobj, bucket, key, *args, **kwargs):
        self.mock_extras.upload_fileobj(*args, **kwargs)
        self.buckets[bucket][key] = fileobj.read()

    def has_object(self, bucket, key):
        return bucket in self.buckets and key in self.buckets[bucket]

    def _get_byte_stream(self, bucket, key):
        return io.BytesIO(self.buckets[bucket][key])

    def download_file(self, Bucket, Key, Filename, *args, **kwargs):
        self.mock_extras.download_file(*args, **kwargs)
        with open(Filename, "wb") as ff:
            ff.write(self._get_byte_stream(Bucket, Key).read())



from dagster import Bool, Field, Noneable, StringSource, check
from dagster.cli.api import ExecuteRunArgs
from dagster.core.launcher import RunLauncher
from dagster.core.storage.pipeline_run import PipelineRun
from dagster.serdes import ConfigurableClass, ConfigurableClassData, serialize_dagster_namedtuple

from .client import ECSClient


class ECSRunLauncher(RunLauncher, ConfigurableClass):
    """
    This class is the RunLauncher for ECS tasks

    Args:
        key_id (Optional[str]): the AWS access key ID to use
        access_key (Optional[str]): the AWS access key going with that key_id
        family (Optional[str]): what task family you want this task revising
        containername (Optional[str]):  what you want the container the task is on to be called
        imagename (Optional[str]): the URI for the docker image
        memory (Optional[str]): how much memory (in MB) the task needs
        cpu (Optional[str]): how much CPU (in vCPU) the task needs
        inst_data (Optional[ConfigurableClassData]): any data specific to the instance
        region_name (Optional[str]):  which AWS region you're running on
        launch_type (Optional[str]): whether to use EC2 or FARGATE for the task
        grab_logs (Optional[Bool]): Whether to grab logs from ECS


    TODO:  get logs from ECSclient into event log storage
    """

    def __init__(
        self,
        inst_data,
        key_id,
        access_key,
        family,
        containername,
        imagename,
        memory,
        cpu,
        region_name,
        launch_type,
        grab_logs,
    ):
        self._inst_data = check.opt_inst_param(inst_data, "inst_data", ConfigurableClassData)
        self.run_id_to_task_offset = dict()
        self.client = ECSClient(
            region_name=region_name,
            key_id=key_id,
            access_key=access_key,
            launch_type=launch_type,
            grab_logs=grab_logs,
        )
        self.client.set_and_register_task(
            [],
            ["dagster"],
            family=family,
            containername=containername,
            imagename=imagename,
            memory=memory,
            cpu=cpu,
        )

    @classmethod
    def config_type(cls):
        return {
            "key_id": Field(
                Noneable(StringSource),
                is_required=False,
                default_value=None,
                description="the AWS access key ID to use, overriding environment vars",
            ),
            "access_key": Field(
                Noneable(StringSource),
                is_required=False,
                default_value=None,
                description="the AWS access key to use, overriding environment vars",
            ),
            "family": Field(
                StringSource,
                is_required=False,
                default_value="dagstertask",
                description="what family of tasks you want to be revising",
            ),
            "containername": Field(
                StringSource,
                is_required=False,
                default_value="dagstercontainer",
                description="what you want the docker container running the tasks to be called",
            ),
            "imagename": Field(
                StringSource,
                is_required=False,
                default_value="httpd:2.4",
                description="the URI for the docker image of the container",
            ),
            "memory": Field(
                StringSource,
                is_required=False,
                default_value="512",
                description="the memory in MB that the task will need",
            ),
            "cpu": Field(
                StringSource,
                is_required=False,
                default_value="256",
                description="the CPU in VCPU that the task will need",
            ),
            "region_name": Field(
                StringSource,
                is_required=False,
                default_value="us-east-2",
                description="which region the AWS cluster is on",
            ),
            "launch_type": Field(
                StringSource,
                is_required=False,
                default_value="FARGATE",
                description="whether to use EC2 or FARGATE for running the task -- currently only Fargate is supported",
            ),
            "grab_logs": Field(
                Bool,
                is_required=False,
                default_value=True,
                description="whether to pull down ECS logs for completed tasks",
            ),
        }

    @classmethod
    def from_config_value(cls, inst_data, config_value):
        return cls(inst_data=inst_data, **config_value)

    @property
    def inst_data(self):
        return self._inst_data

    def launch_run(self, instance, run, external_pipeline):
        check.inst_param(run, "run", PipelineRun)
        input_json = serialize_dagster_namedtuple(
            ExecuteRunArgs(
                pipeline_origin=external_pipeline.get_origin(),
                pipeline_run_id=run.run_id,
                instance_ref=instance.get_ref(),
            )
        )
        # this maps run configuration to task overrides
        # this way we can pass in parameters from the dagit configuration that user has entered in the UI
        overrides = self.generate_task_overrides(run)
        self.client.run_task(
            command=["api", "execute_run_with_structured_logs", input_json], overrides=overrides,
        )
        self.run_id_to_task_offset[run.run_id] = self.client.offset

    def generate_task_overrides(self, run):
        return {
            "containerOverrides": [
                {
                    "name": self.client.container_name,
                    "environment": [
                        {"name": "PIPELINE", "value": run.pipeline_name},
                        # possibly pull from run.run_config['solids'] other data that user might entered in the UI
                    ],
                },
            ],
        }

    def can_terminate(self, run_id):
        check.str_param(run_id, "run_id")
        return False

    def terminate(self, run_id):
        check.str_param(run_id, "run_id")
        check.not_implemented("Termination not yet implemented")



import os

import boto3
import pytest
from dagster.core.test_utils import instance_for_test
from dagster_aws.ecs import client
from moto import mock_ecs, mock_logs


@mock_ecs
@mock_logs
def test_set_cluster():
    log_client = boto3.client("logs", region_name="us-east-2")
    ecs_client = boto3.client("ecs", region_name="us-east-2")
    ecs_client.create_cluster(clusterName="test_clust")
    testclient = client.ECSClient(
        key_id="Testing",
        access_key="Testing",
        starter_client=ecs_client,
        starter_log_client=log_client,
    )
    assert "test_clust" in testclient.cluster


@mock_ecs
@mock_logs
def test_command():
    log_client = boto3.client("logs", region_name="us-east-2")
    ecs_client = boto3.client("ecs", region_name="us-east-2")
    ecs_client.create_cluster(clusterName="test_ecs_cluster")
    testclient = client.ECSClient(
        key_id="Testing",
        access_key="Testing",
        starter_client=ecs_client,
        starter_log_client=log_client,
    )
    testclient.set_and_register_task(
        ["echoes"], [""], family=" ",
    )
    taskdef = ecs_client.describe_task_definition(taskDefinition=testclient.task_definition_arn)[
        "taskDefinition"
    ]
    assert taskdef["containerDefinitions"][0]["command"][0] == "echoes"


@mock_ecs
@mock_logs
def test_entry():
    log_client = boto3.client("logs", region_name="us-east-2")
    ecs_client = boto3.client("ecs", region_name="us-east-2")
    ecs_client.create_cluster(clusterName="test_ecs_cluster")
    testclient = client.ECSClient(
        key_id="Testing",
        access_key="Testing",
        starter_client=ecs_client,
        starter_log_client=log_client,
    )
    testclient.set_and_register_task(
        [" "], ["entries"], family=" ",
    )
    taskdef = ecs_client.describe_task_definition(taskDefinition=testclient.task_definition_arn)[
        "taskDefinition"
    ]
    assert taskdef["containerDefinitions"][0]["entryPoint"][0] == "entries"


@mock_ecs
@mock_logs
def test_family():
    log_client = boto3.client("logs", region_name="us-east-2")
    ecs_client = boto3.client("ecs", region_name="us-east-2")
    ecs_client.create_cluster(clusterName="test_ecs_cluster")
    testclient = client.ECSClient(
        key_id="Testing",
        access_key="Testing",
        starter_client=ecs_client,
        starter_log_client=log_client,
    )
    testclient.set_and_register_task(
        ["echoes"], [""], family="basefam",
    )
    taskdef = ecs_client.describe_task_definition(taskDefinition=testclient.task_definition_arn)[
        "taskDefinition"
    ]
    assert taskdef["family"] == "basefam"
    logger = taskdef["containerDefinitions"][0]["logConfiguration"]["options"]
    assert "basefam" in logger["awslogs-group"]


@mock_ecs
@mock_logs
def test_region():
    log_client = boto3.client("logs", region_name="us-east-2")
    ecs_client = boto3.client("ecs", region_name="us-east-2")
    ecs_client.create_cluster(clusterName="test_ecs_cluster")
    testclient = client.ECSClient(
        key_id="Testing",
        access_key="Testing",
        starter_client=ecs_client,
        starter_log_client=log_client,
    )
    testclient.set_and_register_task(
        ["echoes"], [""], family="basefam",
    )
    taskdef = ecs_client.describe_task_definition(taskDefinition=testclient.task_definition_arn)[
        "taskDefinition"
    ]
    assert taskdef["family"] == "basefam"
    logger = taskdef["containerDefinitions"][0]["logConfiguration"]["options"]
    assert "us-east-2" in logger["awslogs-region"]


@pytest.mark.skipif(
    "AWS_ECS_TEST_DO_IT_LIVE" not in os.environ,
    reason="This test is slow and requires a live ECS cluster; run only upon explicit request",
)
def test_full_run():
    testclient = client.ECSClient()
    testclient.set_and_register_task(
        ["echo start"], ["/bin/bash", "-c"], family="multimessage",
    )
    networkConfiguration = {
        "awsvpcConfiguration": {
            "subnets": ["subnet-0f3b1467",],
            "securityGroups": ["sg-08627da7435350fa6",],
            "assignPublicIp": "ENABLED",
        }
    }
    testclient.run_task(networkConfiguration=networkConfiguration)

    testclient.set_and_register_task(
        ["echo middle"], ["/bin/bash", "-c"], family="multimessage",
    )
    testclient.run_task(networkConfiguration=networkConfiguration)
    testclient.set_and_register_task(
        ["echo $TEST"], ["/bin/bash", "-c"], family="multimessage",
    )
    testclient.run_task(networkConfiguration=networkConfiguration)
    testclient.spin_all()
    assert testclient.logs_messages == {2: ["end"], 1: ["middle"], 0: ["start"]}


@pytest.mark.skipif(
    "AWS_ECS_TEST_DO_IT_LIVE" not in os.environ,
    reason="This test is slow and requires a live ECS cluster; run only upon explicit request",
)
def test_one_run():
    testclient = client.ECSClient()
    testclient.set_and_register_task(
        ["echo start"], ["/bin/bash", "-c"], family="multimessage",
    )
    networkConfiguration = {
        "awsvpcConfiguration": {
            "subnets": ["subnet-0f3b1467",],
            "securityGroups": ["sg-08627da7435350fa6",],
            "assignPublicIp": "ENABLED",
        }
    }
    testclient.run_task(networkConfiguration=networkConfiguration)
    testclient.spin_all()
    assert testclient.logs_messages == {0: ["start"]}


def test_ecs_run_launcher_inits():
    with instance_for_test(
        overrides={
            "run_launcher": {"module": "dagster_aws.ecs.launcher", "class": "ECSRunLauncher"}
        }
    ):
        pass



import time

import boto3
import botocore.exceptions
from dagster import check


class ECSClient:
    """
    This class governs everything you need to generate, run, and track a specific task on ECS

    Args:
        region_name (Optional[str]):  which AWS region you're running on
        key_id (Optional[str]): the AWS access key ID to use
        access_key (Optional[str]): the AWS access key going with that key_id
        launch_type (Optional[str]): whether to use EC2 or FARGATE for the task
        grab_logs (Optional[bool]):  whether to pull down ECS logs
        starter_client (boto3.client): starter ecs client, used for mocking
        starter_log_client (boto3.client): starter log client, used for mocking


    TODO: find ECS specific logs, if they exist -- right now logs are stdout and stderr
    """

    def __init__(
        self,
        region_name="us-east-2",
        key_id="",
        access_key="",
        launch_type="FARGATE",
        starter_client=None,
        starter_log_client=None,
        grab_logs=True,
    ):
        self.key_id = check.opt_str_param(key_id, "key_id", "")
        self.region_name = check.str_param(region_name, "region_name")
        self.access_key = check.opt_str_param(access_key, "access_key")
        self.launch_type = check.str_param(launch_type, "launch_type")
        self.account_id = "123412341234"
        self.grab_logs = check.bool_param(grab_logs, "grab_logs")
        self.task_definition_arn = None
        self.tasks = []
        self.tasks_done = []
        self.offset = -1
        self.container_name = None
        self.ecs_logs = dict()
        self.logs_messages = dict()
        self.warnings = []
        self.cluster = None
        if starter_client is None:
            self.ecs_client = self.make_client("ecs")
            self.account_id = boto3.client("sts").get_caller_identity().get("Account")
        else:
            self.ecs_client = starter_client
        if starter_log_client is None:
            self.log_client = self.make_client("logs")
        else:
            self.log_client = starter_log_client
        self.set_cluster()

    def make_client(self, client_type, **kwargs):
        """
        constructs an ECS client object with the given params

        Args:
            client_type (str): which ECS client you're trying to produce
            **kwargs: any add'l keyword params
        Returns:
            client: the constructed client object
        """
        if self.key_id != "":
            clients = boto3.client(
                client_type,
                aws_access_key_id=self.key_id,
                aws_secret_access_key=self.access_key,
                region_name=self.region_name,
                **kwargs
            )
        else:
            clients = boto3.client(client_type, region_name=self.region_name, **kwargs)

        return clients

    def set_cluster(self):
        """
        sets the task's compute cluster to be the first available/default compute cluster.
        """
        self.cluster = self.ecs_client.list_clusters()["clusterArns"][0]

    def set_and_register_task(
        self,
        command,
        entrypoint,
        family="echostart",
        containername="basictest",
        imagename="httpd:2.4",
        memory="512",
        cpu="256",
    ):
        """
        Generates a task corresponding to a given docker image, command, and entry point
        while setting instance attrs accordingly (logging and task attrs).
        Args:
            command (List[str]): Command to run on container
            entrypoint (List[str]): Entrypoint for container
            family (Optional[str]): what task family you want this task revising
            containername (Optional[str]):  what you want the container the task is on to be called
            imagename (Optional[str]): the URI for the docker image
            memory (Optional[str]): how much memory (in MB) the task needs
            cpu (Optional[str]): how much CPU (in vCPU) the task needs
        """
        fmtstr = family + str(self.offset + 1)
        try:
            self.log_client.create_log_group(logGroupName="/ecs/{}".format(fmtstr))
        except botocore.exceptions.ClientError as e:
            self.warnings.append(str(e))
        basedict = {
            "executionRoleArn": "arn:aws:iam::{}:role/ecsTaskExecutionRole".format(self.account_id),
            "containerDefinitions": [
                {
                    "logConfiguration": {
                        "logDriver": "awslogs",
                        "options": {
                            "awslogs-group": "/ecs/{}".format(fmtstr),
                            "awslogs-region": "us-east-2",
                            "awslogs-stream-prefix": "ecs",
                        },
                    },
                    "entryPoint": entrypoint,
                    "portMappings": [{"hostPort": 8080, "protocol": "tcp", "containerPort": 8080}],
                    "command": command,
                    "cpu": 0,
                    "environment": [{"name": "TEST", "value": "end"}],
                    "image": imagename,
                    "name": containername,
                }
            ],
            "placementConstraints": [],
            "memory": memory,
            "family": family,
            "requiresCompatibilities": ["FARGATE"],
            "networkMode": "awsvpc",
            "cpu": cpu,
        }
        response = self.ecs_client.register_task_definition(**basedict)
        arn = response["taskDefinition"]["taskDefinitionArn"]
        self.task_definition_arn = arn
        self.container_name = containername

    def run_task(self, **kwargs):
        """
        sets this client's task running.

        TODO: explicitly add overrides as params here for cpu, command, and environment vars

        Args:
            **kwargs: Any add'l params you want the task to have
        """
        taskdict = self.ecs_client.run_task(
            taskDefinition=self.task_definition_arn,
            launchType=self.launch_type,
            cluster=self.cluster,
            **kwargs
        )
        task = taskdict["tasks"][0]
        container = task["containers"][0]
        running_task_arn = container["taskArn"]
        self.tasks.append(running_task_arn)
        self.tasks_done.append(False)
        self.offset += 1

    def check_if_done(self, offset=0):
        """
        polls ECS to see if the task is complete, pulling down logs if so

        TODO:  handle pagination for 100+ tasks
        Args:
             offset (int): which task to check for

        Returns:
             boolean corresponding to completion
        """
        if self.tasks_done[offset]:
            return True
        response = self.ecs_client.describe_tasks(tasks=self.tasks, cluster=self.cluster)
        resp = response["tasks"][offset]
        if resp["lastStatus"] == "STOPPED":
            if self.grab_logs:
                check_arn = resp["taskDefinitionArn"]
                taskDefinition = self.ecs_client.describe_task_definition(taskDefinition=check_arn)[
                    "taskDefinition"
                ]
                logconfig = taskDefinition["containerDefinitions"][0]["logConfiguration"]["options"]
                streamobj = self.log_client.describe_log_streams(
                    logGroupName=logconfig["awslogs-group"],
                    logStreamNamePrefix=logconfig["awslogs-stream-prefix"],
                )
                logs = self.log_client.get_log_events(
                    logGroupName=logconfig["awslogs-group"],
                    logStreamName=streamobj["logStreams"][-1]["logStreamName"],
                )
                self.ecs_logs[offset] = logs["events"]
                self.logs_messages[offset] = [log["message"] for log in logs["events"]]
            self.tasks_done[offset] = True
            return True
        return False

    def spin_til_done(self, offset=0, time_delay=5):
        """
        spinpolls the ECS servers to test if the task is done

        Args:
            offset (int): the offset of the task within the task array
            time_delay (int): how long to wait between polls, seconds
        """
        while not self.check_if_done(offset=offset):
            time.sleep(time_delay)

    def spin_all(self, time_delay=5):
        """
        spinpolls all generated tasks
        Args:
            time_delay (int): how long to wait between polls, seconds
        """
        for i in range(self.offset + 1):
            self.spin_til_done(offset=i, time_delay=time_delay)


if __name__ == "__main__":
    import yaml

    with open("config.yaml") as f:
        data = yaml.load(f, Loader=yaml.SafeLoader)

    testclient = ECSClient()
    testclient.set_and_register_task(
        ["echo start"], ["/bin/bash", "-c"], family="multimessage",
    )
    networkConfiguration = {
        "awsvpcConfiguration": {
            "subnets": ["subnet-0f3b1467",],
            "securityGroups": ["sg-08627da7435350fa6",],
            "assignPublicIp": "ENABLED",
        }
    }
    testclient.run_task(networkConfiguration=networkConfiguration)

    testclient.set_and_register_task(
        ["echo middle"], ["/bin/bash", "-c"], family="multimessage",
    )
    testclient.run_task(networkConfiguration=networkConfiguration)
    testclient.set_and_register_task(
        ["echo $TEST"], ["/bin/bash", "-c"], family="multimessage",
    )
    testclient.run_task(networkConfiguration=networkConfiguration)
    testclient.spin_til_done(offset=2)
    testclient.spin_til_done(offset=1)
    testclient.spin_til_done()
    print(testclient.logs_messages)  # pylint: disable=print-call



from dagster_aws.emr.utils import subset_run_config


def test_subset_run_config():
    run_config = {
        "solids": {"blah": {"config": {"foo": "a string", "bar": 123}}},
        "resources": {
            "pyspark": {
                "config": {
                    "pipeline_file": "dagster_aws_tests/emr_tests/test_pyspark.py",
                    "pipeline_fn_name": "pipe",
                    "cluster_id": "j-272P42200OZ0Q",
                    "staging_bucket": "dagster-scratch-80542c2",
                    "region_name": "us-west-1",
                }
            }
        },
    }
    res = subset_run_config(run_config, "blah")
    assert res == run_config

    res = subset_run_config(run_config, "not_here")
    assert res["solids"] == {}



# Portions of this file are copied from the Yelp MRJob project:
#
#   https://github.com/Yelp/mrjob
#
#
# Copyright 2009-2013 Yelp, David Marin
# Copyright 2015 Yelp
# Copyright 2017 Yelp
# Copyright 2018 Contributors
# Copyright 2019 Yelp and Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import gzip
import re
from io import BytesIO

import boto3
import dagster
import six
from botocore.exceptions import WaiterError
from dagster import check
from dagster.seven import urlparse
from dagster_aws.utils.mrjob.utils import _boto3_now, _wrap_aws_client, strip_microseconds

from .types import EMR_CLUSTER_TERMINATED_STATES, EmrClusterState, EmrStepState

# if we can't create or find our own service role, use the one
# created by the AWS console and CLI
_FALLBACK_SERVICE_ROLE = "EMR_DefaultRole"

# if we can't create or find our own instance profile, use the one
# created by the AWS console and CLI
_FALLBACK_INSTANCE_PROFILE = "EMR_EC2_DefaultRole"


class EmrError(Exception):
    pass


class EmrJobRunner:
    def __init__(
        self, region, check_cluster_every=30, aws_access_key_id=None, aws_secret_access_key=None,
    ):
        """This object encapsulates various utilities for interacting with EMR clusters and invoking
        steps (jobs) on them.

        See also :py:class:`~dagster_aws.emr.EmrPySparkResource`, which wraps this job runner in a
        resource for pyspark workloads.

        Args:
            region (str): AWS region to use
            check_cluster_every (int, optional): How frequently to poll boto3 APIs for updates.
                Defaults to 30 seconds.
            aws_access_key_id ([type], optional): AWS access key ID. Defaults to None, which will
                use the default boto3 credentials chain.
            aws_secret_access_key ([type], optional): AWS secret access key. Defaults to None, which
                will use the default boto3 credentials chain.
        """
        self.region = check.str_param(region, "region")

        # This is in seconds
        self.check_cluster_every = check.int_param(check_cluster_every, "check_cluster_every")
        self.aws_access_key_id = check.opt_str_param(aws_access_key_id, "aws_access_key_id")
        self.aws_secret_access_key = check.opt_str_param(
            aws_secret_access_key, "aws_secret_access_key"
        )

    def make_emr_client(self):
        """Creates a boto3 EMR client. Construction is wrapped in retries in case client connection
        fails transiently.

        Returns:
            botocore.client.EMR: An EMR client
        """
        raw_emr_client = boto3.client(
            "emr",
            aws_access_key_id=self.aws_access_key_id,
            aws_secret_access_key=self.aws_secret_access_key,
            region_name=self.region,
        )
        return _wrap_aws_client(raw_emr_client, min_backoff=self.check_cluster_every)

    def cluster_id_from_name(self, cluster_name):
        """Get a cluster ID in the format "j-123ABC123ABC1" given a cluster name "my cool cluster".

        Args:
            cluster_name (str): The name of the cluster for which to find an ID

        Returns:
            str: The ID of the cluster

        Raises:
            EmrError: No cluster with the specified name exists
        """
        check.str_param(cluster_name, "cluster_name")

        response = self.make_emr_client().list_clusters().get("Clusters", [])
        for cluster in response:
            if cluster["Name"] == cluster_name:
                return cluster["Id"]

        raise EmrError(
            "cluster {cluster_name} not found in region {region}".format(
                cluster_name=cluster_name, region=self.region
            )
        )

    @staticmethod
    def construct_step_dict_for_command(step_name, command, action_on_failure="CONTINUE"):
        """Construct an EMR step definition which uses command-runner.jar to execute a shell command
        on the EMR master.

        Args:
            step_name (str): The name of the EMR step (will show up in the EMR UI)
            command (str): The shell command to execute with command-runner.jar
            action_on_failure (str, optional): Configure action on failure (e.g., continue, or
                terminate the cluster). Defaults to 'CONTINUE'.

        Returns:
            dict: Step definition dict
        """
        check.str_param(step_name, "step_name")
        check.list_param(command, "command", of_type=str)
        check.str_param(action_on_failure, "action_on_failure")

        return {
            "Name": step_name,
            "ActionOnFailure": action_on_failure,
            "HadoopJarStep": {"Jar": "command-runner.jar", "Args": command},
        }

    def add_tags(self, log, tags, cluster_id):
        """Add tags in the dict tags to cluster cluster_id.

        Args:
            log (DagsterLogManager): Log manager, for logging
            tags (dict): Dictionary of {'key': 'value'} tags
            cluster_id (str): The ID of the cluster to tag
        """
        check.dict_param(tags, "tags")
        check.str_param(cluster_id, "cluster_id")

        tags_items = sorted(tags.items())

        self.make_emr_client().add_tags(
            ResourceId=cluster_id, Tags=[dict(Key=k, Value=v) for k, v in tags_items]
        )

        log.info(
            "Added EMR tags to cluster %s: %s"
            % (cluster_id, ", ".join("%s=%s" % (tag, value) for tag, value in tags_items))
        )

    def run_job_flow(self, log, cluster_config):
        """Create an empty cluster on EMR, and return the ID of that job flow.

        Args:
            log (DagsterLogManager): Log manager, for logging
            cluster_config (dict): Configuration for this EMR job flow. See:
                https://docs.aws.amazon.com/emr/latest/APIReference/API_RunJobFlow.html

        Returns:
            str: The cluster ID, e.g. "j-ZKIY4CKQRX72"
        """
        check.dict_param(cluster_config, "cluster_config")

        log.debug("Creating Elastic MapReduce cluster")
        emr_client = self.make_emr_client()

        log.debug(
            "Calling run_job_flow(%s)"
            % (", ".join("%s=%r" % (k, v) for k, v in sorted(cluster_config.items())))
        )
        cluster_id = emr_client.run_job_flow(**cluster_config)["JobFlowId"]

        log.info("Created new cluster %s" % cluster_id)

        # set EMR tags for the cluster
        tags = cluster_config.get("Tags", {})
        tags["__dagster_version"] = dagster.__version__
        self.add_tags(log, tags, cluster_id)
        return cluster_id

    def describe_cluster(self, cluster_id):
        """Thin wrapper over boto3 describe_cluster.

        Args:
            cluster_id (str): Cluster to inspect

        Returns:
            dict: The cluster info. See:
                https://docs.aws.amazon.com/emr/latest/APIReference/API_DescribeCluster.html
        """
        check.str_param(cluster_id, "cluster_id")

        emr_client = self.make_emr_client()
        return emr_client.describe_cluster(ClusterId=cluster_id)

    def describe_step(self, cluster_id, step_id):
        """Thin wrapper over boto3 describe_step.

        Args:
            cluster_id (str): Cluster to inspect
            step_id (str): Step ID to describe

        Returns:
            dict: The step info. See:
                https://docs.aws.amazon.com/emr/latest/APIReference/API_DescribeStep.html
        """
        check.str_param(cluster_id, "cluster_id")
        check.str_param(step_id, "step_id")

        emr_client = self.make_emr_client()
        return emr_client.describe_step(ClusterId=cluster_id, StepId=step_id)

    def add_job_flow_steps(self, log, cluster_id, step_defs):
        """Submit the constructed job flow steps to EMR for execution.

        Args:
            log (DagsterLogManager): Log manager, for logging
            cluster_id (str): The ID of the cluster
            step_defs (List[dict]): List of steps; see also `construct_step_dict_for_command`

        Returns:
            List[str]: list of step IDs.
        """
        check.str_param(cluster_id, "cluster_id")
        check.list_param(step_defs, "step_defs", of_type=dict)

        emr_client = self.make_emr_client()

        steps_kwargs = dict(JobFlowId=cluster_id, Steps=step_defs)
        log.debug(
            "Calling add_job_flow_steps(%s)"
            % ",".join(("%s=%r" % (k, v)) for k, v in steps_kwargs.items())
        )
        return emr_client.add_job_flow_steps(**steps_kwargs)["StepIds"]

    def is_emr_step_complete(self, log, cluster_id, emr_step_id):
        step = self.describe_step(cluster_id, emr_step_id)["Step"]
        step_state = EmrStepState(step["Status"]["State"])

        if step_state == EmrStepState.Pending:
            cluster = self.describe_cluster(cluster_id)["Cluster"]

            reason = _get_reason(cluster)
            reason_desc = (": %s" % reason) if reason else ""

            log.info("PENDING (cluster is %s%s)" % (cluster["Status"]["State"], reason_desc))
            return False

        elif step_state == EmrStepState.Running:
            time_running_desc = ""

            start = step["Status"]["Timeline"].get("StartDateTime")
            if start:
                time_running_desc = " for %s" % strip_microseconds(_boto3_now() - start)

            log.info("RUNNING%s" % time_running_desc)
            return False

        # we're done, will return at the end of this
        elif step_state == EmrStepState.Completed:
            log.info("COMPLETED")
            return True
        else:
            # step has failed somehow. *reason* seems to only be set
            # when job is cancelled (e.g. 'Job terminated')
            reason = _get_reason(step)
            reason_desc = (" (%s)" % reason) if reason else ""

            log.info("%s%s" % (step_state.value, reason_desc))

            # print cluster status; this might give more context
            # why step didn't succeed
            cluster = self.describe_cluster(cluster_id)["Cluster"]
            reason = _get_reason(cluster)
            reason_desc = (": %s" % reason) if reason else ""
            log.info(
                "Cluster %s %s %s%s"
                % (
                    cluster["Id"],
                    "was" if "ED" in cluster["Status"]["State"] else "is",
                    cluster["Status"]["State"],
                    reason_desc,
                )
            )

            if EmrClusterState(cluster["Status"]["State"]) in EMR_CLUSTER_TERMINATED_STATES:
                # was it caused by IAM roles?
                self._check_for_missing_default_iam_roles(log, cluster)

                # TODO: extract logs here to surface failure reason
                # See: https://github.com/dagster-io/dagster/issues/1954

        if step_state == EmrStepState.Failed:
            log.error("EMR step %s failed" % emr_step_id)

        raise EmrError("EMR step %s failed" % emr_step_id)

    def _check_for_missing_default_iam_roles(self, log, cluster):
        """If cluster couldn't start due to missing IAM roles, tell user what to do."""

        check.dict_param(cluster, "cluster")

        reason = _get_reason(cluster)
        if any(
            reason.endswith("/%s is invalid" % role)
            for role in (_FALLBACK_INSTANCE_PROFILE, _FALLBACK_SERVICE_ROLE)
        ):
            log.warning(
                "IAM roles are missing. See documentation for IAM roles on EMR here: "
                "https://docs.aws.amazon.com/emr/latest/ManagementGuide/emr-iam-roles.html"
            )

    def log_location_for_cluster(self, cluster_id):
        """EMR clusters are typically launched with S3 logging configured. This method inspects a
        cluster using boto3 describe_cluster to retrieve the log URI.

        Args:
            cluster_id (str): The cluster to inspect.

        Raises:
            EmrError: the log URI was missing (S3 log mirroring not enabled for this cluster)

        Returns:
            (str, str): log bucket and key
        """
        check.str_param(cluster_id, "cluster_id")

        # The S3 log URI is specified per job flow (cluster)
        log_uri = self.describe_cluster(cluster_id)["Cluster"].get("LogUri", None)

        # ugh, seriously boto3?! This will come back as string "None"
        if log_uri == "None" or log_uri is None:
            raise EmrError("Log URI not specified, cannot retrieve step execution logs")

        # For some reason the API returns an s3n:// protocol log URI instead of s3://
        log_uri = re.sub("^s3n", "s3", log_uri)
        log_uri_parsed = urlparse(log_uri)
        log_bucket = log_uri_parsed.netloc
        log_key_prefix = log_uri_parsed.path.lstrip("/")
        return log_bucket, log_key_prefix

    def retrieve_logs_for_step_id(self, log, cluster_id, step_id):
        """Retrieves stdout and stderr logs for the given step ID.

        Args:
            log (DagsterLogManager): Log manager, for logging
            cluster_id (str): EMR cluster ID
            step_id (str): EMR step ID for the job that was submitted.

        Returns
            (str, str): Tuple of stdout log string contents, and stderr log string contents
        """
        check.str_param(cluster_id, "cluster_id")
        check.str_param(step_id, "step_id")

        log_bucket, log_key_prefix = self.log_location_for_cluster(cluster_id)

        prefix = "{log_key_prefix}{cluster_id}/steps/{step_id}".format(
            log_key_prefix=log_key_prefix, cluster_id=cluster_id, step_id=step_id
        )
        stdout_log = self.wait_for_log(log, log_bucket, "{prefix}/stdout.gz".format(prefix=prefix))
        stderr_log = self.wait_for_log(log, log_bucket, "{prefix}/stderr.gz".format(prefix=prefix))
        return stdout_log, stderr_log

    def wait_for_log(self, log, log_bucket, log_key, waiter_delay=30, waiter_max_attempts=20):
        """Wait for gzipped EMR logs to appear on S3. Note that EMR syncs logs to S3 every 5
        minutes, so this may take a long time.

        Args:
            log_bucket (str): S3 bucket where log is expected to appear
            log_key (str): S3 key for the log file
            waiter_delay (int): How long to wait between attempts to check S3 for the log file
            waiter_max_attempts (int): Number of attempts before giving up on waiting

        Raises:
            EmrError: Raised if we waited the full duration and the logs did not appear

        Returns:
            str: contents of the log file
        """
        check.str_param(log_bucket, "log_bucket")
        check.str_param(log_key, "log_key")
        check.int_param(waiter_delay, "waiter_delay")
        check.int_param(waiter_max_attempts, "waiter_max_attempts")

        log.info(
            "Attempting to get log: s3://{log_bucket}/{log_key}".format(
                log_bucket=log_bucket, log_key=log_key
            )
        )

        s3 = _wrap_aws_client(boto3.client("s3"), min_backoff=self.check_cluster_every)
        waiter = s3.get_waiter("object_exists")
        try:
            waiter.wait(
                Bucket=log_bucket,
                Key=log_key,
                WaiterConfig={"Delay": waiter_delay, "MaxAttempts": waiter_max_attempts},
            )
        except WaiterError as err:
            six.raise_from(
                EmrError("EMR log file did not appear on S3 after waiting"), err,
            )
        obj = BytesIO(s3.get_object(Bucket=log_bucket, Key=log_key)["Body"].read())
        gzip_file = gzip.GzipFile(fileobj=obj)
        return gzip_file.read().decode("utf-8")


def _get_reason(cluster_or_step):
    """Get state change reason message."""
    # StateChangeReason is {} before the first state change
    return cluster_or_step["Status"]["StateChangeReason"].get("Message", "")



import pytest


@pytest.fixture(scope="session")
def emr_cluster_config():
    return {
        "Name": "test-emr",
        "LogUri": "s3n://emr-cluster-logs/elasticmapreduce/",
        "ReleaseLabel": "emr-5.23.0",
        "Instances": {
            "MasterInstanceType": "m4.large",
            "SlaveInstanceType": "m4.large",
            "InstanceCount": 4,
            "TerminationProtected": False,
        },
        "Applications": [{"Name": "Spark"}],
        "BootstrapActions": [
            {
                "Name": "Spark Default Config",
                "ScriptBootstrapAction": {
                    "Path": "s3://support.elasticmapreduce/spark/maximize-spark-default-config"
                },
            }
        ],
        "Steps": [
            {
                "Name": "Setup Debugging",
                "ActionOnFailure": "TERMINATE_CLUSTER",
                "HadoopJarStep": {"Jar": "command-runner.jar", "Args": ["state-pusher-script"]},
            },
        ],
        "VisibleToAllUsers": True,
        "JobFlowRole": "EMR_EC2_DefaultRole",
        "ServiceRole": "EMR_DefaultRole",
    }



import io

from dagster_aws.s3 import S3FileCache, S3FileHandle


def test_s3_file_cache_file_not_present(s3, bucket):
    file_store = S3FileCache(
        s3_bucket=bucket.name, s3_key="some-key", s3_session=s3.meta.client, overwrite=False
    )

    assert not file_store.has_file_object("foo")


def test_s3_file_cache_file_present(s3, bucket):
    file_store = S3FileCache(
        s3_bucket=bucket.name, s3_key="some-key", s3_session=s3.meta.client, overwrite=False
    )

    assert not file_store.has_file_object("foo")

    file_store.write_binary_data("foo", "bar".encode())

    assert file_store.has_file_object("foo")


def test_s3_file_cache_correct_handle(s3, bucket):
    file_store = S3FileCache(
        s3_bucket=bucket.name, s3_key="some-key", s3_session=s3.meta.client, overwrite=False
    )

    assert isinstance(file_store.get_file_handle("foo"), S3FileHandle)


def test_s3_file_cache_write_file_object(s3, bucket):
    file_store = S3FileCache(
        s3_bucket=bucket.name, s3_key="some-key", s3_session=s3.meta.client, overwrite=False
    )

    stream = io.BytesIO("content".encode())
    file_store.write_file_object("foo", stream)



from dagster import ModeDefinition, execute_pipeline, pipeline, solid
from dagster.utils.test import get_temp_file_handle_with_data
from dagster_aws.s3 import file_handle_to_s3, s3_resource


def create_file_handle_pipeline(temp_file_handle):
    @solid
    def emit_temp_handle(_):
        return temp_file_handle

    @pipeline(mode_defs=[ModeDefinition(resource_defs={"s3": s3_resource})])
    def test():
        return file_handle_to_s3(emit_temp_handle())

    return test


def test_successful_file_handle_to_s3(bucket):
    foo_bytes = "foo".encode()
    with get_temp_file_handle_with_data(foo_bytes) as temp_file_handle:
        result = execute_pipeline(
            create_file_handle_pipeline(temp_file_handle),
            run_config={
                "solids": {
                    "file_handle_to_s3": {"config": {"Bucket": bucket.name, "Key": "some-key"}}
                }
            },
        )

        assert result.success

        assert bucket.Object(key="some-key").get()["Body"].read() == foo_bytes

        materializations = result.result_for_solid(
            "file_handle_to_s3"
        ).materializations_during_compute
        assert len(materializations) == 1
        assert len(materializations[0].metadata_entries) == 1
        assert materializations[0].metadata_entries[
            0
        ].entry_data.path == "s3://{bucket}/some-key".format(bucket=bucket.name)
        assert materializations[0].metadata_entries[0].label == "some-key"
